#!/usr/bin/env python
################################################
# setenv PYTHONPATH ./orgasm/build/lib.macosx-10.6-intel-2.7/
# echo $PYTHONPATH
# ./orgasm/build/lib.macosx-10.6-intel-2.7/
# python orgasmb.py ./samples/GWM-261 ./samples/COX1_Machaon.fasta machaon

from orgasm.tango import tango, \
                         matchtoseed, \
                         cutLowCoverage, \
                         scaffold, \
                         coverageEstimate, \
                         fillGaps, \
                         cutSNPs, \
                         estimateDeadBrancheLength,\
                         estimateFragmentLength, \
                         genesincontig
from orgasm.assembler import Assembler
from orgasm.tango import dumpGraph

import orgasm.samples
                           
import sys
from orgasm.config import getConfiguration

from orgasm import getProbes, getIndex
from orgasm import getSeeds

def estimateMinRead(index,minoverlap,coverage):
    MINREAD=10
    MINREADCOR=3
    MINOVERLAP=50
    minread =  (index.getReadSize() - minoverlap) * coverage / index.getReadSize()  / MINREADCOR
    if minread < MINREAD:
        minoverlap = index.getReadSize() - (MINREAD * MINREADCOR * index.getReadSize() / coverage)
        minread = MINREAD
    if  minoverlap< MINOVERLAP:
        minread =  MINREAD
        minoverlap = MINOVERLAP
    return minread,minoverlap

if __name__ == '__main__':
    
    import argparse
    
    # Declare script options
    
    parser = argparse.ArgumentParser(description='Assembly program dedicated to organelle genome reconstruction.')
    parser.add_argument('indexFilename', metavar='index', help='index root filename (produced by orgasmi)')
    parser.add_argument('outputFilename', metavar='output', help='output prefix', nargs='?', default=None)
    
    parser.add_argument('--minread', dest='minread', type=int, action='store', default=None, help='the minimum count of read to consider [default: <estimated>]')
    parser.add_argument('--coverage', dest='coverage', type=int, action='store', default=None, help='the expected sequencing coverage [default: <estimated>]')
    parser.add_argument('--minratio', dest='minratio', type=float, action='store', default=None, help='minimum ratio between occurrences of an extension and the occurrences of the most frequent extension to keep it. [default: <estimated>]')
    parser.add_argument('--mincov', dest='mincov', type=int, action='store', default=1, help='minimum occurrences of an extension to keep it. [default: %(default)d]')
    parser.add_argument('--minoverlap', dest='minoverlap', type=int, action='store', default=None, help='minimum length of the overlap between the sequence and reads to participate in the extension. [default: <estimated>]')
    parser.add_argument('--smallbranches', dest='smallbranches', type=int, action='store', default=None, help='maximum length of the branches to cut during the cleaning process [default: <estimated>]')
    parser.add_argument('--lowcomplexity', dest='lowcomplexity', action='store_true', default=False, help='Use also low complexity probes')
    parser.add_argument('--back', dest='back', type=int, action='store', default=None, help='the number of bases taken at the end of contigs to jump with pared-ends [default: <estimated>]')
    parser.add_argument('--snp', dest='snp', action='store_true', default=False, help='desactivate the SNP clearing mode')
    parser.add_argument('--log', dest='log', action='store_true', default=False, help='Create a logfile for the assembling')
    parser.add_argument('--seeds', dest ='seeds', metavar='seeds', default=None, type=str, help='protein seeds; either a file containing seeds proteic sequences or internal set of seeds among %s' % str(filter(lambda s: s.startswith('prot') or s.startswith('nuc'),dir(orgasm.samples))) )
    #parser.add_argument('--maxjump', dest='maxjump', type=int, action='store', default=1, help='COunt of pair end jump by extension branch')
    
    
    args = parser.parse_args()
    
    config = getConfiguration(args)
    

    r = getIndex(config)
    p = getProbes(config)
    x = getSeeds(r,p,config)            
            
    # First esti;ation of the coverage based on protein matches
    
    common    = set(p.keys()) & set(x.keys())
    pcoverage = [len(x[i]) * r.getReadSize()  / len(p[i]) / 3 for i in common]
    pcoverage.sort()

    config.logger.info("Probe coverage : %s" % str(pcoverage))

    #coverage=mode(pcoverage)
    #coverage=pcoverage[len(pcoverage)/2]
    if config.coverage is None:
        coverage=pcoverage[-1]
    else:
        coverage = config.coverage
    
    # according to the minread option estimate it from coverage or use the specified value
   
    if config.minread is None:
        minread,minoverlap = estimateMinRead(r,config.minoverlap,coverage)
    else:
        minread = config.minread
        minoverlap = config.minoverlap

    config.logger.info("\nestimated coverage  : %d based on probe match (minread: %d)" % (coverage,minread))
    
        
        
    # Convert matches in sedd list    
    s = matchtoseed(x,r)
    
    # Create the assembler object
    asm = Assembler(r)

    ##########################
    #
    # If minread is not specified we initiate the assembling
    # based on the coverage estimated from protein match
    # to obtain a better coverage estimation
    #
    ##########################
    if config.minread is None:

        # Run the first assembling pass
        a = tango(asm,s,mincov=minread,       # config.mincov,
                        minread=minread,
                        minoverlap=minoverlap,
                        lowfilter=not config.lowcomplexity,
                        maxjump=0, 
                        cycle=1,
                        nodeLimit=30000)
    
        # Clean small unsuccessful extensions
        asm.cleanDeadBranches(maxlength=10)
        
        # and too low covered assembling
        if config.coverage is None:
            cutLowCoverage(asm,int(coverage/4),terminal=True)
        else:
            cutLowCoverage(asm,int(coverage),terminal=True)
            
    
        if config.snp:
            cutSNPs(asm)
        
        if config.smallbranches is not None:
            smallbranches = config.smallbranches
        else:
            smallbranches = estimateDeadBrancheLength(asm)
            config.logger.info("Dead branch length setup to : %d bp" % smallbranches)
            asm.cleanDeadBranches(maxlength=smallbranches)
        
        if len(asm) > 0:
            score,length,ecoverage = coverageEstimate(asm)
            if config.coverage is None:
                coverage = ecoverage    
            minread,minoverlap = estimateMinRead(r,config.minoverlap,coverage) 
            config.logger.info("coverage estimated : %d based on %d bp (minread: %d)" %(coverage,length,minread/2))
        
            
        # Reinitiate the assembling for running with the estimated parameters
        # Convert matches in sedd list    
        s = matchtoseed(x,r)
    
        # Create the assembler object
        asm = Assembler(r)

    #############################################
    #
    # We now run the main assembling process
    #
    #############################################

    # Run the first assembling pass
    a = tango(asm,s,mincov=coverage/4,       #config.mincov,
                    minread=minread,
                    minoverlap=minoverlap,
                    lowfilter=not config.lowcomplexity,
                    maxjump=0, 
                    cycle=1)
    
    # Clean small unsuccessful extensions
    asm.cleanDeadBranches(maxlength=10)
    
    # and too low covered assembling
    if config.coverage is None:
        cutLowCoverage(asm,int(coverage/4),terminal=True)
    else:
        cutLowCoverage(asm,int(coverage),terminal=True)
        
    
    # cleanup snp bubble in the graph    
    if config.snp:
        cutSNPs(asm)
    
    if config.smallbranches is not None:
        smallbranches = config.smallbranches
    else:
        smallbranches = estimateDeadBrancheLength(asm)
        config.logger.info("     Dead branch length setup to : %d bp" % smallbranches)

    asm.cleanDeadBranches(maxlength=smallbranches)


    # reestimate coverage
    score,length,ecoverage = coverageEstimate(asm)  
    
    if config.coverage is None:
        coverage = ecoverage  
    
    if coverage < 30:
        sys.exit()

    if config.minread is None:
        minread,minoverlap = estimateMinRead(r,config.minoverlap,coverage)
        minread/=4
        if minread<5:
            minread=5

    config.logger.info("coverage estimated : %d based on %d bp (minread: %d)" %(coverage,length,minread))
        
    meanlength,sdlength = estimateFragmentLength(asm)
    
    if config.back is not None:
        back = config.back
    elif config.back is None and meanlength is not None:
        back = int(meanlength + 4 * sdlength)
        if back > 500:
            back=500
    else:
        back = 300
        
    if meanlength is not None:
        config.logger.info("Fragment length estimated : %f pb (sd: %f)" % (meanlength,sdlength))

    cg = asm.compactAssembling(verbose=False)
    genesincontig(cg,r,x)
    scaffold(asm,cg,minlink=5,back=back,addConnectedLink=False)
    print >>open(args.outputFilename+'.intermediate.gml','w'),cg.gml()
    
    ###################################################
    #
    # We now fill the gaps between the contigs
    #
    ###################################################
    
    delta = 1
       
    # Run the fill gap procedure    
    while  delta > 0 or delta < -2000 :
                   
        delta = fillGaps(asm,back=back,
                       minread=minread,
                       maxjump=0,
                       minoverlap=minoverlap,
                       cmincov=2,
                       emincov=int(coverage/4),
                       gmincov=int(coverage/4),
                       snp=args.snp)

        print >>sys.stderr
        print >>sys.stderr
        print >>sys.stderr
        print >>sys.stderr,"=================================================================="      
        print >>sys.stderr 
        
        # and too low covered assembling
        # cutLowCoverage(asm,int(coverage/3),terminal=True)
    
    
        ## reestimate coverage
        #score,length,coverage = coverageEstimate(asm)    
        #if config.minread is None:
        #    minread,minoverlap = estimateMinRead(r,config.minoverlap,coverage)
        #    minread/=4
        #    if minread<5:
        #        minread=5
        #
        #config.logger.info("coverage estimated : %d based on %d bp (minread: %d)" %(coverage,length,minread))

        cg = asm.compactAssembling(verbose=False)
        genesincontig(cg,r,x)
        scaffold(asm,cg,minlink=5,back=back,addConnectedLink=False)
        print >>open(config.outputname+'.intermediate.gml','w'),cg.gml()
        
        if meanlength is None:
            meanlength,sdlength = estimateFragmentLength(asm)
            if config.back is None and meanlength is not None:
                config.logger.info("Fragment length estimated : %f pb (sd: %f)" % (meanlength,sdlength))
                back = int(meanlength + 4 * sdlength)  
                if back > 500:
                    back=500
                         
        print >>sys.stderr
        print >>sys.stderr,"=================================================================="      
        print >>sys.stderr
        

    ###################################################
    #
    # Finishing of the assembling
    #
    ###################################################

    if config.snp:
        config.logger.info("Clean polymorphisms")
        cutSNPs(asm)
        
    asi = len(asm)+1
    
    config.logger.info("Clean dead branches")
    while (asi>len(asm)):
        asi=len(asm)
        smallbranches = estimateDeadBrancheLength(asm)
        config.logger.info("     Dead branch length setup to : %d bp" % smallbranches)
        asm.cleanDeadBranches(maxlength=smallbranches)
        
    cg = asm.compactAssembling(verbose=False)
    score,length,ecoverage = coverageEstimate(asm)
    if config.coverage is None:
        coverage=ecoverage

    if config.snp:
        config.logger.info("Clean polymorphisms phase 2")
        cutSNPs(asm)
        
    config.logger.info("Clean low coverage terminal branches")
    if config.coverage is None:
        cutLowCoverage(asm,int(coverage/2),terminal=True)
        config.logger.info("Clean low coverage internal branches")
        cutLowCoverage(asm,int(coverage/3),terminal=False)
    else:
        cutLowCoverage(asm,int(coverage),terminal=False)
        
        
    cg = asm.compactAssembling(verbose=False)     
        
    config.logger.info("Scaffold the assembly")
    scaffold(asm,cg,minlink=5,back=int(back),addConnectedLink=False)
    genesincontig(cg,r,x)
    print >>open(config.outputname+'.gml','w'),cg.gml()
    dumpGraph(config.outputname+'.oax',asm)
    
